from typing import Iterator, Tuple

from event_sourcery.dto.raw_event_dict import RawEventDict
from event_sourcery.interfaces.outbox_storage_strategy import (
    EntryId,
    OutboxStorageStrategy,
)


class DummyOutboxStorageStrategy(OutboxStorageStrategy):
    def put_into_outbox(self, events: list[RawEventDict]) -> None:
        pass

    def outbox_entries(self, limit: int) -> Iterator[Tuple[EntryId, RawEventDict]]:
        return iter([])

    def decrease_tries_left(self, entry_id: EntryId) -> None:
        pass

    def remove_from_outbox(self, entry_id: EntryId) -> None:
        pass



import abc
from typing import Type

from event_sourcery.dto.raw_event_dict import RawEventDict
from event_sourcery.interfaces.event import Event
from event_sourcery.types.stream_id import StreamId


class Serde(abc.ABC):
    @abc.abstractmethod
    def deserialize(self, event: RawEventDict, event_type: Type[Event]) -> Event:
        pass

    @abc.abstractmethod
    def serialize(
        self, event: Event, stream_id: StreamId, name: str, version: int
    ) -> RawEventDict:
        pass



from dataclasses import dataclass
from typing import Callable, Iterator, Union

from sqlalchemy import delete
from sqlalchemy import event as sa_event
from sqlalchemy import insert, select, update
from sqlalchemy.dialects.postgresql import insert as postgresql_insert
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import Session

from event_sourcery.dto.raw_event_dict import RawEventDict
from event_sourcery.exceptions import ConcurrentStreamWriteError
from event_sourcery.interfaces.storage_strategy import StorageStrategy
from event_sourcery.types.stream_id import StreamId
from event_sourcery_sqlalchemy.models import Event as EventModel
from event_sourcery_sqlalchemy.models import Snapshot as SnapshotModel
from event_sourcery_sqlalchemy.models import Stream as StreamModel


@dataclass(repr=False)
class SqlAlchemyStorageStrategy(StorageStrategy):
    _session: Session

    def iter(self, *stream_ids: StreamId) -> Iterator[RawEventDict]:
        events_stmt = select(EventModel).order_by(EventModel.id).limit(100)
        if stream_ids:
            events_stmt = events_stmt.filter(EventModel.stream_id.in_(stream_ids))

        last_id = 0
        while True:
            stmt = events_stmt.filter(EventModel.id > last_id)
            events = self._session.execute(stmt).scalars().all()
            if not events:
                break
            last_id = events[-1].id
            for event in events:
                yield RawEventDict(
                    uuid=event.uuid,
                    stream_id=event.stream_id,
                    created_at=event.created_at,
                    version=event.version,
                    name=event.name,
                    data=event.data,
                    metadata=event.event_metadata,
                )

    def fetch_events(
        self, stream_id: StreamId, start: int | None = None, stop: int | None = None
    ) -> list[RawEventDict]:
        events_stmt = (
            select(EventModel)
            .filter(EventModel.stream_id == stream_id)
            .order_by(EventModel.version)
        )
        if start is not None:
            events_stmt = events_stmt.filter(EventModel.version >= start)

        if stop is not None:
            events_stmt = events_stmt.filter(EventModel.version < stop)

        events: list[Union[EventModel, SnapshotModel]]
        try:
            snapshot_stmt = (
                select(SnapshotModel)
                .filter(SnapshotModel.stream_id == stream_id)
                .order_by(SnapshotModel.created_at.desc())
                .limit(1)
            )
            if start is not None:
                snapshot_stmt = snapshot_stmt.filter(SnapshotModel.version >= start)

            if stop is not None:
                snapshot_stmt = snapshot_stmt.filter(SnapshotModel.version < stop)

            latest_snapshot = self._session.execute(snapshot_stmt).scalars().one()
        except NoResultFound:
            events = self._session.execute(events_stmt).scalars().all()
        else:
            events_stmt = events_stmt.filter(
                EventModel.version > latest_snapshot.version
            )
            events = [latest_snapshot] + self._session.execute(
                events_stmt
            ).scalars().all()

        if not events:
            return []

        raw_dict_events = [
            RawEventDict(
                uuid=event.uuid,
                stream_id=event.stream_id,
                created_at=event.created_at,
                version=event.version,
                name=event.name,
                data=event.data,
                metadata=event.event_metadata,
            )
            for event in events
        ]
        return raw_dict_events

    def ensure_stream(self, stream_id: StreamId, expected_version: int) -> None:
        ensure_stream_stmt = (
            postgresql_insert(StreamModel)
            .values(
                uuid=stream_id,
                version=1,
            )
            .on_conflict_do_nothing()
        )
        self._session.execute(ensure_stream_stmt)

        if expected_version:
            stmt = (
                update(StreamModel)
                .where(
                    StreamModel.uuid == stream_id,
                    StreamModel.version == expected_version,
                )
                .values(version=StreamModel.version + 1)
            )
            result = self._session.execute(stmt)

            if result.rowcount != 1:  # optimistic lock failed
                raise ConcurrentStreamWriteError

    def insert_events(self, events: list[RawEventDict]) -> None:
        rows = []
        for event in events:
            rows.append(
                {
                    "uuid": event["uuid"],
                    "stream_id": event["stream_id"],
                    "created_at": event["created_at"],
                    "name": event["name"],
                    "data": event["data"],
                    "event_metadata": event["metadata"],
                    "version": event["version"],
                }
            )

        self._session.execute(insert(EventModel), rows)

    def save_snapshot(self, snapshot: RawEventDict) -> None:
        snapshot_as_dict = dict(snapshot)
        row = {
            "uuid": snapshot_as_dict.pop("uuid"),
            "stream_id": snapshot_as_dict.pop("stream_id"),
            "created_at": snapshot_as_dict.pop("created_at"),
            "version": snapshot_as_dict.pop("version"),
            "name": snapshot_as_dict.pop("name"),
            "data": snapshot_as_dict["data"],
            "event_metadata": snapshot_as_dict["metadata"],
        }
        self._session.execute(insert(SnapshotModel), [row])

    def delete_stream(self, stream_id: StreamId) -> None:
        delete_events_stmt = delete(EventModel).where(
            EventModel.stream_id == str(stream_id)
        )
        self._session.execute(delete_events_stmt)
        delete_stream_stmt = delete(StreamModel).where(
            StreamModel.uuid == str(stream_id)
        )
        self._session.execute(delete_stream_stmt)

    def run_after_commit(self, callback: Callable[[], None]) -> None:
        sa_event.listen(self._session, "after_commit", lambda _session: callback())



import json
from typing import Mapping, Type, cast

from event_sourcery.dto.raw_event_dict import RawEventDict
from event_sourcery.interfaces.event import Event
from event_sourcery.interfaces.serde import Serde
from event_sourcery.types.stream_id import StreamId
from event_sourcery_pydantic.event import Event as PydanticEvent


class PydanticSerde(Serde):
    def deserialize(self, event: RawEventDict, event_type: Type[Event]) -> Event:
        event_as_dict = dict(event)
        data = cast(Mapping, event_as_dict.pop("data"))
        return cast(Event, event_type(**event_as_dict, **data))

    def serialize(
        self, event: Event, stream_id: StreamId, name: str, version: int
    ) -> RawEventDict:
        model = cast(PydanticEvent, event)
        as_dict = json.loads(  # json dumps and loads? It's moronic
            model.json(exclude={"uuid", "created_at", "version"})
        )
        metadata = as_dict.pop("metadata")
        return RawEventDict(
            uuid=model.uuid,
            stream_id=stream_id,
            created_at=model.created_at,
            version=model.version or version,
            name=name,
            data=as_dict,
            metadata=metadata,
        )



import abc
from typing import Iterator, Sequence, Type, TypeVar

from event_sourcery.after_commit_subscriber import AfterCommit
from event_sourcery.dto.raw_event_dict import RawEventDict
from event_sourcery.event_registry import BaseEventCls, EventRegistry
from event_sourcery.exceptions import Misconfiguration, NoEventsToAppend
from event_sourcery.interfaces.event import Event
from event_sourcery.interfaces.outbox_storage_strategy import OutboxStorageStrategy
from event_sourcery.interfaces.serde import Serde
from event_sourcery.interfaces.storage_strategy import StorageStrategy
from event_sourcery.interfaces.subscriber import Subscriber
from event_sourcery.types.stream_id import StreamId

TAggregate = TypeVar("TAggregate")


class EventStore(abc.ABC):
    def __init__(
        self,
        serde: Serde,
        storage_strategy: StorageStrategy,
        outbox_storage_strategy: OutboxStorageStrategy,
        event_base_class: Type[BaseEventCls] | None = None,
        event_registry: EventRegistry | None = None,
        subscriptions: dict[Type[Event], list[Subscriber]] | None = None,
    ) -> None:
        if event_base_class is not None and event_registry is not None:
            raise Misconfiguration(
                "You can specify only one of `event_base_class` or `event_registry`"
            )

        if subscriptions is None:
            subscriptions = {}

        if event_base_class is not None:
            self._event_registry = event_base_class.__registry__
        elif event_registry is not None:
            self._event_registry = event_registry
        else:  # not possible
            pass

        self._serde = serde
        self._storage_strategy = storage_strategy
        self._outbox_storage_strategy = outbox_storage_strategy
        self._subscriptions = subscriptions

    def load_stream(
        self, stream_id: StreamId, start: int | None = None, stop: int | None = None
    ) -> list[Event]:
        events = self._storage_strategy.fetch_events(stream_id, start=start, stop=stop)
        return self._deserialize_events(events)

    def append(
        self, stream_id: StreamId, events: Sequence[Event], expected_version: int = 0
    ) -> None:
        if not events:
            raise NoEventsToAppend

        self._storage_strategy.ensure_stream(
            stream_id=stream_id, expected_version=expected_version
        )

        serialized_events = self._serialize_events(events, stream_id, expected_version)
        self._storage_strategy.insert_events(serialized_events)

        # TODO: make it more robust per subscriber?
        for event in events:
            # Shall we support async subscribers here?
            # Shall we support after-commit-subscribers?
            for subscriber in self._subscriptions.get(type(event), []):
                if isinstance(subscriber, AfterCommit):
                    self._storage_strategy.run_after_commit(lambda: subscriber(event))
                else:
                    subscriber(event)

            catch_all_subscribers = self._subscriptions.get(Event, [])  # type: ignore
            for catch_all_subscriber in catch_all_subscribers:
                catch_all_subscriber(event)

    def publish(
        self, stream_id: StreamId, events: Sequence[Event], expected_version: int = 0
    ) -> None:
        self.append(
            stream_id=stream_id, events=events, expected_version=expected_version
        )
        # this happens twice, should be optimized away
        serialized_events = self._serialize_events(events, stream_id, expected_version)

        # Rethink. Should it be always putting into outbox
        # once we have async subscribers?
        self._outbox_storage_strategy.put_into_outbox(serialized_events)

    def iter(self, *streams_ids: StreamId) -> Iterator[Event]:
        events_iterator = self._storage_strategy.iter(*streams_ids)
        for event in events_iterator:
            yield self._serde.deserialize(
                event, self._event_registry.type_for_name(event["name"])
            )

    def delete_stream(self, stream_id: StreamId) -> None:
        self._storage_strategy.delete_stream(stream_id)

    def save_snapshot(self, stream_id: StreamId, snapshot: Event, version: int) -> None:
        serialized = self._serde.serialize(
            event=snapshot,
            stream_id=stream_id,
            name=self._event_registry.name_for_type(type(snapshot)),
            version=version,
        )
        self._storage_strategy.save_snapshot(serialized)

    def _deserialize_events(self, events: list[RawEventDict]) -> list[Event]:
        return [
            self._serde.deserialize(
                event=event,
                event_type=self._event_registry.type_for_name(event["name"]),
            )
            for event in events
        ]

    def _serialize_events(
        self, events: Sequence[Event], stream_id: StreamId, expected_stream_version: int
    ) -> list[RawEventDict]:
        return [
            self._serde.serialize(
                event=event,
                stream_id=stream_id,
                name=self._event_registry.name_for_type(type(event)),
                version=version,
            )
            for version, event in enumerate(events, start=expected_stream_version + 1)
        ]



from typing import ClassVar, Protocol, Type

from event_sourcery.interfaces.event import Event


class EventRegistry:
    def __init__(self) -> None:
        self._types_to_names: dict[Type[Event], str] = {}
        self._names_to_types: dict[str, Type[Event]] = {}

    def add(self, event: Type[Event]) -> Type[Event]:
        if event.__name__ in self._types_to_names.values():
            raise Exception(f"Duplicated Event name detected! {event.__name__}")

        self._types_to_names[event] = event.__name__
        self._names_to_types[event.__name__] = event
        return event  # for use as a decorator

    def type_for_name(self, name: str) -> Type[Event]:
        return self._names_to_types[name]

    def name_for_type(self, event: Type[Event]) -> str:
        return self._types_to_names[event]


class BaseEventCls(Protocol):
    __registry__: ClassVar[EventRegistry]



import abc
from typing import Callable, Type

from event_sourcery.event_registry import BaseEventCls
from event_sourcery.interfaces.event import Event
from event_sourcery.interfaces.outbox_storage_strategy import OutboxStorageStrategy
from event_sourcery.interfaces.serde import Serde


class Outbox(abc.ABC):
    CHUNK_SIZE = 100

    def __init__(
        self,
        serde: Serde,
        storage_strategy: OutboxStorageStrategy,
        event_base_class: Type[BaseEventCls],
        publisher: Callable[[Event], None],
    ) -> None:
        self._serde = serde
        self._storage_strategy = storage_strategy
        self._event_registry = event_base_class.__registry__
        self._publisher = publisher

    def run_once(self) -> None:
        raw_event_dicts = self._storage_strategy.outbox_entries(limit=self.CHUNK_SIZE)
        for entry_id, raw_event_dict in raw_event_dicts:
            event = self._serde.deserialize(
                event=raw_event_dict,
                event_type=self._event_registry.type_for_name(raw_event_dict["name"]),
            )
            try:
                self._publisher(event)
            except Exception:
                self._storage_strategy.decrease_tries_left(entry_id)
            else:
                self._storage_strategy.remove_from_outbox(entry_id)



from event_sourcery.interfaces.event import Event
from event_sourcery.interfaces.subscriber import Subscriber


class AfterCommit(Subscriber):
    def __init__(self, subscriber: Subscriber) -> None:
        self._subscriber = subscriber

    def __call__(self, event: Event) -> None:
        self._subscriber(event)



from datetime import datetime
from typing import TypedDict
from uuid import UUID


class RawEventDict(TypedDict):
    uuid: UUID
    stream_id: UUID
    created_at: datetime
    version: int
    name: str
    data: dict
    metadata: dict



from datetime import datetime
from typing import Any, ClassVar, Optional
from uuid import UUID, uuid4

from pydantic import BaseModel, Extra, Field

from event_sourcery.event_registry import EventRegistry
from event_sourcery.interfaces.event import AUTO_VERSION


class Metadata(BaseModel, extra=Extra.allow):
    correlation_id: Optional[UUID]
    causation_id: Optional[UUID]


class Event(BaseModel):
    uuid: UUID = Field(default_factory=uuid4)
    created_at: datetime = Field(default_factory=datetime.utcnow)
    version: int = AUTO_VERSION
    metadata: Metadata = Field(default_factory=Metadata)

    __registry__: ClassVar = EventRegistry()

    def __init_subclass__(cls, **kwargs: Any) -> None:
        cls.__registry__.add(cls)



import json
from typing import Any, Optional

from sqlalchemy.dialects.postgresql import JSONB as PostgresJSONB
from sqlalchemy.types import Text, TypeDecorator


class JSONB(TypeDecorator):
    impl = Text

    def load_dialect_impl(self, dialect: Any) -> Any:
        if dialect.name == "postgresql":
            return dialect.type_descriptor(PostgresJSONB())
        else:
            return dialect.type_descriptor(Text())

    def process_bind_param(self, value: Any, dialect: Any) -> Optional[Any]:
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return value
        else:
            return json.dumps(value)

    def process_result_value(self, value: Any, dialect: Any) -> Optional[Any]:
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return value
        else:
            return json.loads(value)



from typing import Type

from sqlalchemy import BigInteger, Column, DateTime, Index, Integer, String
from sqlalchemy.ext.declarative import instrument_declarative

from event_sourcery_sqlalchemy.guid import GUID
from event_sourcery_sqlalchemy.jsonb import JSONB


def configure_models(base: Type) -> None:
    for model_cls in (Stream, Event, Snapshot, OutboxEntry):
        instrument_declarative(model_cls, {}, base.metadata)


class Stream:
    __tablename__ = "event_sourcery_streams"

    uuid = Column(GUID(), primary_key=True)
    version = Column(BigInteger(), nullable=False)


class Event:
    __tablename__ = "event_sourcery_events"
    __table_args__ = (
        Index(
            "ix_events_stream_id_version",
            "stream_id",
            "version",
            unique=True,
        ),
    )

    id = Column(BigInteger().with_variant(Integer(), "sqlite"), primary_key=True)
    version = Column(Integer(), nullable=False)
    uuid = Column(GUID(), index=True, unique=True)
    stream_id = Column(GUID(), nullable=False, index=True)
    name = Column(String(50), nullable=False)
    data = Column(JSONB(), nullable=False)
    event_metadata = Column(JSONB(), nullable=False)
    created_at = Column(DateTime(), nullable=False, index=True)


class Snapshot:
    __tablename__ = "event_sourcery_snapshots"

    uuid = Column(GUID, primary_key=True)
    version = Column(Integer(), nullable=False)
    stream_id = Column(GUID(), nullable=False, index=True)
    name = Column(String(50), nullable=False)
    data = Column(JSONB(), nullable=False)
    event_metadata = Column(JSONB(), nullable=False)
    created_at = Column(DateTime(), nullable=False)


class OutboxEntry:
    __tablename__ = "event_sourcery_outbox_entries"

    id = Column(BigInteger().with_variant(Integer(), "sqlite"), primary_key=True)
    created_at = Column(DateTime(), nullable=False, index=True)
    data = Column(JSONB(), nullable=False)
    tries_left = Column(Integer(), nullable=False, server_default="3")



from dataclasses import dataclass
from datetime import datetime
from typing import Iterator, Tuple, cast

from sqlalchemy import insert, select
from sqlalchemy.orm import Session

from event_sourcery.dto.raw_event_dict import RawEventDict
from event_sourcery.interfaces.outbox_storage_strategy import (
    EntryId,
    OutboxStorageStrategy,
)
from event_sourcery_sqlalchemy.models import OutboxEntry


@dataclass(repr=False)
class SqlAlchemyOutboxStorageStrategy(OutboxStorageStrategy):
    _session: Session

    def put_into_outbox(self, events: list[RawEventDict]) -> None:
        rows = []
        for event in events:
            as_dict = dict(event)
            created_at = cast(datetime, as_dict["created_at"])
            as_dict["created_at"] = created_at.isoformat()
            as_dict["uuid"] = str(as_dict["uuid"])
            as_dict["stream_id"] = str(as_dict["stream_id"])
            rows.append(
                {
                    "created_at": datetime.utcnow(),
                    "data": as_dict,
                }
            )
        self._session.execute(insert(OutboxEntry), rows)

    def outbox_entries(self, limit: int) -> Iterator[Tuple[EntryId, RawEventDict]]:
        stmt = (
            select(OutboxEntry)
            .filter(OutboxEntry.tries_left > 0)
            .order_by(OutboxEntry.id)
            .limit(limit)
            .with_for_update(skip_locked=True)
        )
        entries = self._session.execute(stmt).scalars().all()
        return ((entry.id, entry.data) for entry in entries)

    def decrease_tries_left(self, entry_id: EntryId) -> None:
        entry = self._session.query(OutboxEntry).get(entry_id)
        entry.tries_left -= 1

    def remove_from_outbox(self, entry_id: EntryId) -> None:
        entry = self._session.query(OutboxEntry).get(entry_id)
        self._session.delete(entry)



class EventStoreException(Exception):
    pass


class NoEventsToAppend(EventStoreException):
    pass


class ConcurrentStreamWriteError(EventStoreException):
    pass


class Misconfiguration(EventStoreException):
    pass



from contextlib import contextmanager
from typing import Any, Iterator, Type

from event_sourcery.interfaces.event import Event


class Aggregate:
    def __init__(self) -> None:
        self.__version = 0
        self.__changes: list[Event] = []

    def __rehydrate__(self, event: Event) -> None:
        self._apply(event)
        self.__version = event.version

    @contextmanager
    def __persisting_changes__(self) -> Iterator[Iterator[Event]]:
        yield iter(self.__changes)
        self.__version = self.__changes[-1].version
        self.__changes = []

    @property
    def __version__(self) -> int:
        return self.__version

    def _apply(self, event: Event) -> None:
        raise NotImplementedError()

    def _event(self, event_cls: Type[Event], **kwargs: Any) -> None:
        if self.__changes:
            next_version = self.__changes[-1].version + 1
        else:
            next_version = self.__version + 1

        event = event_cls(version=next_version, **kwargs)
        self._apply(event)
        self.__changes.append(event)



import abc
from typing import Iterator, Tuple

from event_sourcery.dto.raw_event_dict import RawEventDict

EntryId = int


class OutboxStorageStrategy(abc.ABC):
    @abc.abstractmethod
    def put_into_outbox(self, events: list[RawEventDict]) -> None:
        pass

    @abc.abstractmethod
    def outbox_entries(self, limit: int) -> Iterator[Tuple[EntryId, RawEventDict]]:
        pass

    @abc.abstractmethod
    def decrease_tries_left(self, entry_id: EntryId) -> None:
        pass

    @abc.abstractmethod
    def remove_from_outbox(self, entry_id: EntryId) -> None:
        pass



__all__ = [
    "configure_models",
    "get_event_store",
    "Event",
    "get_outbox",
    "Repository",
    "Aggregate",
]

from typing import Callable, Type

from sqlalchemy.orm import Session

from event_sourcery.aggregate import Aggregate
from event_sourcery.dummy_outbox_storage_strategy import DummyOutboxStorageStrategy
from event_sourcery.event_store import EventStore
from event_sourcery.interfaces.event import Event as EventProtocol
from event_sourcery.interfaces.outbox_storage_strategy import OutboxStorageStrategy
from event_sourcery.interfaces.subscriber import Subscriber
from event_sourcery.outbox import Outbox
from event_sourcery.repository import Repository
from event_sourcery_pydantic.event import Event
from event_sourcery_pydantic.serde import PydanticSerde
from event_sourcery_sqlalchemy.models import configure_models
from event_sourcery_sqlalchemy.sqlalchemy_event_store import SqlAlchemyStorageStrategy
from event_sourcery_sqlalchemy.sqlalchemy_outbox import SqlAlchemyOutboxStorageStrategy


def get_event_store(
    session: Session,
    subscriptions: dict[Type[EventProtocol], list[Subscriber]] | None = None,
    with_outbox: bool = True,
) -> EventStore:
    outbox_storage: OutboxStorageStrategy
    if with_outbox:
        outbox_storage = SqlAlchemyOutboxStorageStrategy(session)
    else:
        outbox_storage = DummyOutboxStorageStrategy()

    return EventStore(
        serde=PydanticSerde(),
        storage_strategy=SqlAlchemyStorageStrategy(session),
        outbox_storage_strategy=outbox_storage,
        event_base_class=Event,
        subscriptions=subscriptions,
    )


def get_outbox(session: Session, publisher: Callable[[EventProtocol], None]) -> Outbox:
    return Outbox(
        serde=PydanticSerde(),
        storage_strategy=SqlAlchemyOutboxStorageStrategy(session),
        event_base_class=Event,
        publisher=publisher,
    )



from contextlib import contextmanager
from typing import Generic, Iterator, Type, TypeVar

from event_sourcery.aggregate import Aggregate
from event_sourcery.event_store import EventStore
from event_sourcery.types.stream_id import StreamId

TAggregate = TypeVar("TAggregate", bound=Aggregate)


class Repository(Generic[TAggregate]):
    def __init__(
        self, event_store: EventStore, aggregate_cls: Type[TAggregate]
    ) -> None:
        self._event_store = event_store
        self._aggregate_cls = aggregate_cls

    class NotFound(Exception):
        pass

    def load(self, stream_id: StreamId) -> TAggregate:
        aggregate = object.__new__(self._aggregate_cls)
        Aggregate.__init__(aggregate)
        events = self._event_store.load_stream(stream_id)
        for event in events:
            aggregate.__rehydrate__(event)
        return aggregate

    def save(self, aggregate: TAggregate, stream_id: StreamId) -> None:
        with aggregate.__persisting_changes__() as events:
            self._event_store.append(
                stream_id=stream_id,
                events=list(events),
                expected_version=aggregate.__version__,
            )

    @contextmanager
    def aggregate(self, stream_id: StreamId) -> Iterator[TAggregate]:
        aggregate = self.load(stream_id)
        yield aggregate
        self.save(aggregate, stream_id)

    def load_up(self, aggregate: TAggregate, stream_id: StreamId) -> None:
        # protocode, would have to implement that ability in EventStore first
        events = [
            event
            for event in self._event_store.iter(stream_id)
            if event.version > aggregate.__version__
        ]
        for event in events:
            aggregate.__rehydrate__(event)
        # TODO: that should fail if aggregate's state is not clean (has pending events)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['event_sourcery',
 'event_sourcery.dto',
 'event_sourcery.interfaces',
 'event_sourcery.types',
 'event_sourcery_pydantic',
 'event_sourcery_sqlalchemy']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4,<2.0', 'pydantic>=1.9,<2.0']

setup_kwargs = {
    'name': 'python-event-sourcery',
    'version': '0.1.6',
    'description': '',
    'long_description': None,
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from datetime import datetime
from typing import Any, Final, Optional, Protocol
from uuid import UUID


class Metadata(Protocol):
    correlation_id: Optional[UUID]
    causation_id: Optional[UUID]


AUTO_VERSION: Final = 0


class Event(Protocol):
    uuid: UUID
    created_at: datetime
    version: int = AUTO_VERSION

    def __init__(self, **kwargs: Any) -> None:
        ...  # pragma: no cover

    @property
    def metadata(self) -> Metadata:
        # https://mypy.readthedocs.io/en/latest/common_issues.html#covariant-subtyping-of-mutable-protocol-members-is-rejected
        ...  # pragma: no cover



import abc
from typing import Callable, Iterator

from event_sourcery.dto.raw_event_dict import RawEventDict
from event_sourcery.types.stream_id import StreamId


class StorageStrategy(abc.ABC):
    @abc.abstractmethod
    def fetch_events(
        self, stream_id: StreamId, start: int | None = None, stop: int | None = None
    ) -> list[RawEventDict]:
        pass

    @abc.abstractmethod
    def iter(self, *stream_ids: StreamId) -> Iterator[RawEventDict]:
        pass

    @abc.abstractmethod
    def insert_events(self, events: list[RawEventDict]) -> None:
        pass

    @abc.abstractmethod
    def save_snapshot(self, snapshot: RawEventDict) -> None:
        pass

    @abc.abstractmethod
    def ensure_stream(self, stream_id: StreamId, expected_version: int) -> None:
        pass

    @abc.abstractmethod
    def delete_stream(self, stream_id: StreamId) -> None:
        pass

    @abc.abstractmethod
    def run_after_commit(self, callback: Callable[[], None]) -> None:
        pass
